var swipeCategories = {
    isDebug: false,
    lazyQueries: [],
    categories: [],
    useCategories: true, 
    defaultCategoryID: null,
    paneWrapper: null,
    delay: 700,
    limit: 30,
    limitFirst: 30,
    currentPane: 0,
    myScroll: new Array(),
    loading: '<li class="txtC"><div class="circular-small"><div class="circularG_1 circularG"></div><div class="circularG_2 circularG"></div><div class="circularG_3 circularG"></div><div class="circularG_4 circularG"></div><div class="circularG_5 circularG"></div><div class="circularG_6 circularG"></div><div class="circularG_7 circularG"></div><div class="circularG_8 circularG"></div></div></li>',
    initialize: function(params) {
        this.useCategories = params.useCategories; 
        this.defaultCategoryID = params.defaultCategoryID;
        
        this.initCategories();
        this.createHtml();

        $(window).unbind('resize.category').on('resize.category', function(e) {
            loading('show');
            swipeCategories.correctSizes(e);
        });

        this.initSwipe();
    },
    
    initCategories: function() {
        var pane = 0;
        var query = getDecodeParameterByName("query");

        if( query !== null ) {
            $("#topCategories ul").append('<li data-pane="0" class="category0">All</li>');
            swipeCategories.categories.push({name: 'All', id: 0});

            $("#categories-scroller > ul").append('<li class="pane pane' + pane + '" data-pane="' + pane + '" data-category_id="0"><div id="wrapper' + pane + '" class="wrapper"><div class="scroller"><ul class="dealsList clearfix"><li></li></ul><a href="#" class="getDealsElement hide">Get More Deals</a></div></div></li>');

            pane++;     
        } else {
            if( swipeCategories.useCategories ) {
                $("#topCategoriesWrap, #triangle-up").show();
            }
        }
        
        $.each(Vocabulary.categories, function(i, el) {
            if( el.id != 4 ) {
                if( query !== null || (swipeCategories.defaultCategoryID !== null && swipeCategories.defaultCategoryID != el.id) ) {
                    return;
                }
                var name = el.name.toString();
                var id = parseInt(el.id);
                swipeCategories.categories.push({name:name, id: id});

                $("#topCategories > ul").append('<li data-id="'+id+'" data-pane="'+pane+'">'+name+'<\/li>');  

                $("#categories-scroller > ul").append('<li style="overflow:hidden;" class="pane pane' + pane + '" data-pane="' + pane + '" data-category_id="' + el.id + '"><div id="wrapper' + pane + '" class="wrapper"><div class="scroller"><ul class="mvn ui-listview dealsList clearfix"><li style="text-shadow: 0px 1px 0px #fff; text-transform: uppercase; padding:100px 0 !important; color:#ccc; text-align:center; font-size:30px">Loading</li></ul><a href="#" class="getDealsElement hide">Get More Deals</a></div></div></li>');

                pane++;
            }
        });

        $("#topCategories ul li.category1").addClass('active');
        
        $("#topCategories ul li").on('click', function() {
            var pane = $(this).data('pane');
            var time = Math.abs(swipeCategories.currentPane - pane) * 300;
        
            swipeCategories.currentPane = pane;
            swipeCategories.focusSelectedCategory();
            
            swipeCategories.lazyLoad(pane, function() {
                setTimeout(function() {
                    loading('hide');
                    swipeCategories.paneWrapper.scrollToPage(pane, 0, time);
                }, 200);
            });
        });
    },
    debug: function(str) {
        if( this.isDebug ) {
            _log('swipeCategories', str);
        }
    },
    lazyLoad: function(pane, callback) {
        if( swipeCategories.lazyQueries[pane] ) {
            swipeCategories.debug('lazy load: make ajax query. load pane #'+pane);
            loading('show');
            swipeCategories.lazyQueries[pane](callback);
            swipeCategories.lazyQueries[pane] = null;
        } else {
            swipeCategories.debug('lazy load: pane '+pane+' is already loaded');
            callback();
        }
        
        var prevPane = pane - 1;
        var nextPane = pane + 1;
       
        if( swipeCategories.lazyQueries[prevPane] && typeof swipeCategories.lazyQueries[prevPane] == 'function' ) { // !== null && swipeCategories.lazyQueries[prevPane] !== undefined
            setTimeout(function() {
                swipeCategories.debug('lazy load: make prev pane ajax query. load pane #'+prevPane);
                
                swipeCategories.lazyQueries[prevPane](); 
                swipeCategories.lazyQueries[prevPane] = null;
            }, 100);  
        }
        
        if( swipeCategories.lazyQueries[nextPane] && typeof swipeCategories.lazyQueries[nextPane] == 'function' ) { // !== null && swipeCategories.lazyQueries[nextPane] !== undefined
            setTimeout(function() {
                swipeCategories.debug('lazy load: make next pane ajax query. load pane #'+nextPane);
                swipeCategories.lazyQueries[nextPane]();
                swipeCategories.lazyQueries[nextPane] = null;  
            }, 200);
            
        }
    },
    correctSizes: function(e) {
        swipeCategories.debug('correctSizes');
        setTimeout(function(){
            var windowWidth = $(window).width();
            var windowHeight = $(window).height();
            var headerHeight = $("#header").outerHeight();
            
            $("#topCategories").width(windowWidth);
            var width = 0;
            $("#topCategories li").each(function(k, v) {
                width += $(v).outerWidth(true) + 1;
            });
            $("#topCategories ul").width(width);

            $("#wrapper").css({top: headerHeight, height:(windowHeight - headerHeight)});
            $(".scroller, .wrapper, .pane").height(windowHeight - headerHeight);

            $("#wrapper, #categories-scroller > ul > li").width(windowWidth);
            $("#categories-scroller").width($("#categories-scroller > ul > li").length * windowWidth);

            if (swipeCategories.paneWrapper) {
                swipeCategories.paneWrapper.refresh();
                swipeCategories.paneWrapper.scrollToPage(swipeCategories.currentPane, 0, 0);
            }

            fixMiddleContainer({
                container: ".scroller"
            });

            swipeCategories.focusSelectedCategory();
            if( e !== undefined && e.type ) {
                loading('hide');
            }
        }, 10);
    },
    initInfiniteScroll: function() {
        swipeCategories.debug('initInfiniteScroll');
        $('.wrapper').each(function() {
            var id = $(this).attr('id');
            
            (function(id) {
                var scroller = $('#' + id).infinitescroll({
                    navSelector: '#' + id + " a.getDealsElement",
                    nextSelector: '#' + id + " a.getDealsElement",
                    itemSelector: '#' + id + " .scroller ul li",
                    
                    debug: false,
                    dataType: 'json',
                    dataType: 'javascript',
                    binder: $('#' + id + ' .scroller'),
                    state: {
                        currPage: 1
                    },
                    behavior: 'custom',
                    loading: {
                        msg: $(swipeCategories.loading),
                        msgText: null,
                        speed: 0,
                        selector: '#' + id + " .scroller ul",
                    },
                    appendCallback: false,
                    path: function(nextPage) {
                        dealsParams.offset = (nextPage - 1) * swipeCategories.limit;
                        dealsParams.limit = swipeCategories.limit;
                        dealsParams.categoryID = $('#' + id).parent().data('category_id');
                        return api.getDealsUrl(dealsParams);
                    }
                }, function( empty, instance ) {
                    instance.state.currentPage = instance.state.currentPage || 2;

                    app.sendRequestCache = function() {
                        instance.state.lock = true;
                        instance.state.isDone = true;

                        _log('app.sendRequestCache');

                        var url = instance.path(instance.state.currentPage);
                        $(instance.navSelector).hide();
                        instance.loading.msg
                                .appendTo(instance.loading.selector)
                                .show(instance.loading.speed);
                        
                        app.sendRequest({
                            log: ['infinitScroll', url],
                            dataType: 'json',
                            type: 'GET',
                            url: url,
                            complete: function() {
                                instance.state.isDone = false;
                                instance.loading.msg.fadeOut(instance.loading.speed);
                            },
                            success: function (response) {
                                app.dataLoading++;

                                getDealsSuccess(response);
                                showList(true, '#' + id + ' .dealsList');
                                
                                instance.state.currentPage++;
                                app.sendRequestCache = null;
                                instance.state.lock = false;
                                if( response.deals && response.deals.length < 1 ) {
                                    instance.state.lock = true;
                                    _log('infinitScroll', 'isBeyondMaxPage');
                                }
                            },
                            error: function() {
                                if( leave == false ) {
                                    instance.state.lock = false;
                                    if( navigator.notification ) {
                                        navigator.notification.confirm(
                                            'Please try again', 
                                            function(buttonIndex){
                                                if( buttonIndex == 1 ) {
                                                    app.sendRequestCache();
                                                }
                                            }, 'We could not reach the server', 
                                            ['Try again','Cancel']
                                        );
                                    } else {
                                        if( confirm('We could not reach the server. Please try again') ) {
                                            app.sendRequestCache();
                                        }
                                    }
                                }
                                return;
                            }
                        });
                        return;
                    };

                    if( instance.state.lock != true && app.sendRequestCache ) {
                        app.sendRequestCache(); 
                    }
                }); 
                swipeCategories.myScroll.push(scroller);
            })(id);
        });
    },
    createHtml: function() {
        swipeCategories.debug('createHtml');
        function makeQuery(options) {
            var num = options.num || 0;
            var delay = options.delay || 0;
            setTimeout(function() {
                dealsParams.limit = swipeCategories.limitFirst;
                dealsParams.offset = 0;
                dealsParams.categoryID = options.categoryID;

                var categoryName = $('#topCategories ul li[data-id='+dealsParams.categoryID+']').text()

                var url = api.getDealsUrl(dealsParams);
                $.ajax({
                    type: 'get',
                    url: url,
                    cache: false,
                    dataType: 'json',
                    success: function(json) {
                        _log('swipeCategories.createHtml', 'success load ' + categoryName + ' deals', url);
                        getDealsSuccess(json);

                        showList(false, '.pane' + num + ' .dealsList');

                        if( options.callback ) {
                            options.callback();
                        }
                    },
                    error: function() {
                        if( leave == false ) {
                            if( navigator.notification ) {
                                navigator.notification.confirm(
                                    'Please try again', 
                                    function(buttonIndex){
                                        if( buttonIndex == 1 ) {
                                            _log('swipeCategories.createHtml', 'try load again', url);
                                            var _options = JSON.parse(JSON.stringify(options));
                                            _options.delay = 0;
                                            makeQuery(_options);
                                        }
                                    }, 
                                    categoryName + '. We could not reach the server', 
                                    ['Try again','Cancel']
                                );
                            } else {
                                if( confirm(categoryName + '. We could not reach the server. Please try again') ) {
                                    _log('swipeCategories.createHtml', 'try load again', url);
                                    var _options = JSON.parse(JSON.stringify(options));
                                    _options.delay = 0;
                                    makeQuery(_options);
                                }
                            }
                        }
                    },
                    timeout: ajaxRequestTimeout
                });
            }, swipeCategories.delay * parseInt(delay));
        }
        
        $.each(swipeCategories.categories, function(i, el) {
            var categoryID = el.id || i;

            if( i < 3 ) {
                var callback = false;

                if (i == 0) {
                    callback = function(){
                        setTimeout(function() {
                            loading('hide');
                        }, 100); 
                    };
                }
                makeQuery({num:i, categoryID:categoryID, delay:i, callback: callback});
                swipeCategories.lazyQueries.push(null);
            } else {
                swipeCategories.lazyQueries.push(function(callback) {
                    makeQuery({num:i, categoryID:categoryID, delay:0, callback:callback});
                });
            }
        });

        this.correctSizes();
        this.initInfiniteScroll();
    },
    focusSelectedCategory: function() {
        currentPane = swipeCategories.currentPane || 0;
        
        swipeCategories.debug('focusSelectedCategory: '+currentPane);
        
        $("#topCategories li").removeClass('active');
        $("#topCategories li[data-pane="+currentPane+"]").addClass('active');

        var scrollLeft = 0;
        $.each($("#topCategories li"), function(k, v) {
            if (k <= currentPane) {
                scrollLeft += $(v).outerWidth(true);
            }
        });

        scrollLeft -= 0 + $("#topCategories li.active").outerWidth(true) * 1;
        if( false ) {
            $("#topCategories ul").css("transform", 'translateX(-' + scrollLeft + 'px)');
        } else {
            $("#topCategories").animate({
                'scrollLeft': scrollLeft
            });
        }
    },
    initSwipe: function() {
        swipeCategories.paneWrapper = new iScroll('wrapper', {
            vScroll: false,
            hScroll: true,
            snap: true,
            snapThreshold: 50,
            handleClick: false,
            momentum: false,
            lockDirection: true,
            bounce: true,
            onBeforeScrollStart: false,
            onScrollMove: function(e) {
                if( this.absDistX > this.absDistY ) {
                    e.preventDefault(); // disable scrolling
                }
            },
            onScrollEnd: function() {
                if( swipeCategories.paneWrapper.enabled ) {
                    var currentPane = this.currPageX;
                    swipeCategories.currentPane = currentPane;
                    swipeCategories.focusSelectedCategory();

                    swipeCategories.lazyLoad(currentPane, function() {
                        loading('hide');
                    });   

                    app.logEvent('App', 'categoryView', 'Select category', {title: $('#topCategories ul li[data-pane='+currentPane+']').text()});
                }
            }
        });
    }
};

(function($, undefined) {
    $.extend($.infinitescroll.prototype, {
        _nearbottom_custom: function() {
            var opts = this.options;
            var pixelsFromWindowBottomToBottom = (opts.binder.find('.dealsList').outerHeight(true) - opts.binder.height()) - opts.binder.scrollTop();
            return pixelsFromWindowBottomToBottom < opts.bufferPx;
        }
    });
})(jQuery);