var userObject = {
    isAuth: false,
    errors: {},
    oldData: {},
    data: {
        id: null,
        name: null,
        lastname: null,
        email: null,
        password: null,
        defaultLocation: null,
        emailNotificationFrequency: null,
        notifyWhenNewDeals: 1,
        alertWhenExpire: 1,
        update: null
    },
    init: function(options) {
        _log('userObject.init');
        
        var self = this;
        var user = getCache('user', 'local');
        if( user ) {
            self.setData(user);
            self.isAuth = true;
        }
        
        return self;
    },
    showErrors: function() {
        $.each(this.errors, function(k, v){
            var input = $('input[name="'+k+'"]');
            
            input.parents('.input-parent').find('.error-text').html(v).show();
            
            input.parent('.input-parent').addClass('error');
            input.parent('.input-wrapper').siblings('.icon-edit').addClass('error');
            input.after('<span class="error-symbol"><i></i></span>');
        });
    },
    addError: function(k, v) {
        this.errors[k] = v;
    },
    validation: function(data) {             
        this.errors = {};

        if( data.name !== undefined ) {
            if( data.name.length < 1 ) {
                this.addError('name', 'The First Name cannot be blank');
            }
        }

        if( data.lastname !== undefined ) {
            if( data.lastname.length < 1 ) {
                this.addError('lastname', 'The Last Name cannot be blank');
            }
        }

        if( data.email !== undefined ) {
            if( data.email.length < 1 ) {
                this.addError('email', 'The Email cannot be blank');
            } else if( /^(([^<>()[\]\\.,;:\s@\"]+(\.[^<>()[\]\\.,;:\s@\"]+)*)|(\".+\"))@((\[[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\])|(([a-zA-Z\-0-9]+\.)+[a-zA-Z]{2,}))$/.test(data.email) !== true ) {
                this.addError('email', 'The Email should be a valid email address');
            }
        }
        
        if( data.password !== undefined ) {
            if( data.password.length < 5 ) {
                this.addError('password', 'Password must be at least 5 characters long');
            }
        }
        return Object.size(this.errors) == 0;
    },
    login: function(email, password) { // аутентификация пользователя
        var status = false;
        var message = 'Success';
        var data = {};

        if( true ) {
            status = true;
            data = {
                id: 5,
                name: 'Alex',
                lastName: 'Chehlov',
                email: 'achehlov@gmail.com',
                password: 'password'
            };

            app.logEvent('User', 'userLogin', 'User Login', {id: data.id, email: data.email});
        } else {
            message = 'Wrong login/password !';
            app.logEvent('User', 'userWrongLogin', 'User tried to login with wrong credentials', {email: email});
        }
        
        _log('userObject.login', 'status:', status, 'message:', message);
        
        return {
            message: message,
            status: status, 
            data: data
        };
    },
    sendData: function(data) {
        var status = false;
        var message = 'Success';
        
        // ....
        
        if( true ) {
            app.logEvent('User', 'userAdd', 'Register new user', {id:data.id, email: data.email});
            
            status = true;
        } else {
            message = 'User already exists';
        }
        
        return {
            message: message,
            status: status
        };   
    },
    auth: function(data) { // авторизация пользователя
        var self = this;
        
        _log('userObject.auth', JSON.stringify(data));
        
        self.setData(data).saveData();
        self.isAuth = true;
        
        return self;
    },
    logout: function() {
        var self = this;

        app.logEvent('User', 'userLogout', 'User Logout', {id: self.data.id, email: self.data.email});
        
        $.each(self.data, function(k, v) {
            self.data[k] = null;
            v = null;
        });
        
        _log('userObject.logout');
        
        removeCache('user', 'local');
    },
    getData: function(options) { 
       
    },
    setData: function(data) {
        var self = this;

        self.oldData = JSON.parse(JSON.stringify(self.data));

        $.each(data, function(k, v) {
            if( self.data[k] !== undefined ) {
                self.data[k] = v;
            }
        });

        return self;
    },   
    saveData: function() {
        var self = this;
        self.data.update = new Date().getTime();
        _log('userObject.saveData', JSON.stringify(self.data));
        saveCache('user', JSON.stringify(self.data), 'local');
        
        if( self.oldData.defaultLocation !== self.data.defaultLocation ) {
            api.setLocationDevice( self.data.defaultLocation );
        }

        if( self.oldData.notifyWhenNewDeals !== self.data.notifyWhenNewDeals ) {
            api.setNewDealsNotification( self.data.notifyWhenNewDeals );
        }
        
        self.oldData = JSON.parse(JSON.stringify(self.data));
        return self;
    },
    getLastUpdate: function() {
        var update = this.data.update || new Date().getTime();
        return moment(update).format('YYYY/DD/MM');
    },
    getDefaultCity: function() {
        var self = this;
        if( self.data.defaultLocation && self.data.defaultLocation != 'all' ) {
            return self.data.defaultLocation;
        }
        return false;
    }
};